/* 

   **********************************************************************
   CLASS:   CGIUpload:  Provides the ability to upload files from
                        a web browser to a web server 

   Version: 4 (01/01/2002)

   Requirements:  Your HTML form must make two important modifications: 
                  First the form tag must set 'enctype = "multipart/form-data'" 
                  Second the input tag must uses type=file.

       <FORM ENCTYPE="multipart/form-data" ACTION="dB2K_App.exe" METHOD=POST>
       File to process: <INPUT NAME="userfile1" TYPE="file">
       <INPUT TYPE="submit" VALUE="Send File">
       </FORM>

   Authors:    Michael Nuwer 
               Bowen Moursund

   **********************************************************************
   CLASS:   CGIDownload:  Provides the ability to download binary or text
                          files from a web server to a web brwowser 
                          
   Authors:    Bowen Moursund
               Michael Nuwer 


   **********************************************************************
   Copyright (C) 2000 Michael Nuwer
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.

   Michael Nuwer <nuwermj@potsdam.edu> 

*/

#define INPUT_BUFFER_SIZE 65536

Class CGIUpload of CGISession from "WebClass.cc"

   // Set the folders where temp file and the uploaded file(s) will be
   // written in the calling PRG. These folders can be anywhere on the server;
   // they do not need to be within "web space". The default is the working
   // folder of the servlet.
   this.TempPath = ""
   this.SavePath = ""

   // Set default WebMasterEMail address
   this["WebMasterEMail"] = ""

   // Override base class Connect function, eliminating call to
   // loadArrayFromCGI. We'll use ReadMultipart instead.
   function Connect

      ////// Open  output stream back to Web Server
      try

         this.fOut = new file()
         this.fOut.Open("StdOut", "RA")  // open the StdOut output stream
         
      catch (exception e)

         quit  // Might as well quit, can't send response

      endTry
      return true


   function ReadMultipart

      local nLen, fIn, cEnv, cBytes, fTemp, cTempName, cString, cEnvBoundry,;
            cBoundry, buffer, cName, cValue, cSaveFile, bSuccess, bMore

      bSuccess = true

      nLen = val(getEnv("CONTENT_LENGTH")) 
      fIn  = new file()           // Create new instance of File Object

      fIn.Open("StdIn", "RWB")  //  Modified for Binary files

      cEnv=""
   
      try 
      ////  There appears to be a bug in Read(), which was identified by Paul Franks.
      ////  This Try ... Endtry and Do ... Until structure is a workaround
      ////  until the bug is fixed -- if infact it is a bug.  
      ////  If the Read() fails we abort.
      ////  Try reading smaller chuncks if you have trouble.
      ////////////////////////////////////////////////////
      ///   This bug was fixed in dB2k-03 (b1636)
      ///   The do loop is no longer necessary
      ////////////////////////////////////////////////////
         do 
        
            cBytes = fIn.Read(INPUT_BUFFER_SIZE)
            cEnv = cEnv + cBytes

         until len(cEnv) = nLen
   
      catch (exception e)

         // Abort! Abort!
         try
            fIn.close()
            fIn = null
         catch (exception e)
         endtry
         return false  
 
      endtry

      try
   
         ///  Create a temp file to store the full CGI environment string.
         fTemp = new file()
         cTempName = funique(this.TempPath + "CGI?????.tmp")
         ftemp.create(cTempName)
         fTemp.write( cEnv )
         fTemp.Seek(0)
   
         cString = ftemp.gets()      ///  Get the first line of the temp file
   
         ///  The CGI data includes a field separator called Boundry
         ///  It is included in  getenv("CONTENT_TYPE")
         ///  and looks something like this:
         ///   -----------------------------16740297823514
         ///  The boundry will be the first line in the temp file.

         cEnvBoundry = cString        ///  Used to control the Do While loop
         cBoundry = cEnvBoundry       ///  Used to Idntify a new CGI field
         buffer = ''

         //  "Walk" down the temp file line-by-line.  The last line in the 
         //  file is Boundry + "--"  Which will drop us out of this loop
         //  To see the structure of the CGI data, don't delete the temp file. 
         //  It can be opened with a text editor.
         do while cEnvBoundry == cBoundry
            do case
            case trim(cString) == trim(cBoundry)
               cString = ftemp.gets()

            case "filename=" $ cString 
               Buffer = ""
               cName = substr(cString,at('name="',cString) + len('name="') )
               cName = substr(cName,1,at('"',cName)-1)
               cValue = substr(cString,at('filename="',cString) + len('filename="') )
               cValue = substr(cValue,1,len(cValue)-1)

               if not empty(cValue)   ///  if there is a file uploaded
                  ///  strip directories from file name
                  do case
                  case at("\",cValue) > 0     ///   WINDOWS=>'\'
                     do while at("\",cValue) > 0 
                        cValue = substr(cValue,at("\",cValue)+1)
                     enddo

                  case at("/",cValue) > 0    ///   UNIX=>'/'
                     do while at("/",cValue) > 0
                        cValue = substr(cValue,at("/",cValue)+1)
                     enddo

                  case at(":",cValue) > 0    ///  MACINTOSH=>':' 
                     do while at(":",cValue) > 0 
                        cValue = substr(cValue,at(":",cValue)+1)
                     enddo

                  endcase

                  this[cName] = cValue
                  cSaveFile = cValue
                  fTemp.gets()
                  fTemp.gets()
         
                  nPosition = fTemp.seek(0,1)
                  bMore = true
                  do while bMore  

                     fByte = fTemp.read(256)
                     buffer = buffer + fByte

                     if cBoundry $ buffer

                        // there is a CR/LF at the end of this line
                        // They should not be included in the binary file

                        buffer = substr(buffer,1,at(cBoundry,buffer)-3)    
  
                        /// We may have read beyond the end for the boundry marker
                        /// We goto the beginning of the boundry.  
                        /// And will read additional form data.
                        fTemp.seek(0)                                   
                        fTemp.seek(nPosition + len(buffer)-1)

                        ///  Write the uploaded file data to the server.
                        fUploadFile = new file()
                        cSaveFile = this.SavePath + cSaveFile
                        fUploadFile.create(cSaveFile)
                        fUploadFile.write(buffer)            
                        fUploadFile.close()
       
                        /// drop out of loop
                        bMore = false
              
                     endif
                  enddo

                  cBoundry = fTemp.Gets()
                  cString = cBoundry

               else

                  // no file was uploaded, 
                  // skip two lines in the temp file   -- If Netscape did the upload
                  // skip three lines                  -- If IE did the upload
                  do                                    /// loop added in version 3

                     cString = fTemp.gets()

                  until substr(cString,1,len(cEnvBoundry)) = cEnvBoundry

               endif


               cBoundry = cString

            case "name=" $ cString
               //  Name value pair
               cName = substr(cString,at('"',cString) + 1)
               cName = substr(cName,1,len(cName)-1)
        
               fTemp.Gets()                /// skip one line in the file


               //cLine = fTemp.Gets()   ///  get the first line of data
               cValue = fTemp.Gets()

               // the following loop is uesed 
               // in case the values are more than one line of text
               // like, for example, data from a text areas.
               cLine = fTemp.Gets()
               do  while substr(cLine,1,len(cEnvBoundry)) <> cBoundry 
                                     /// We use substr()
                                     /// in case we are evaluating
                                     /// the last name/value pair.
                                     /// In that event the CGI boundry
                                     /// is cBoundry + "--"
            
                  cValue = cValue + chr(10) + cLine
                                     ///  If there is more than one line
                                     ///  of text we add a CR for formating.  
                  cLine = fTemp.Gets()
               enddo 

               this[cName] = cValue

               cBoundry = cLine
               cString = cBoundry
       

            otherwise
               /// Just in case 
               fTemp.close()
               fTemp.delete(cTempName)

               bSuccess := false

            endcase

         enddo
         fTemp.close()

         ///  Comment the next line if you what to open 
         ///  the temp file and see the contents of the
         ///  CGI data.

         fTemp.delete(cTempName)

      catch (exception e)

         try
            fTemp.close()
         catch (exception e)
         endtry

         try
            fUploadFile.close()
         catch (exception e)
         endtry

         bSuccess := false

      endtry

      return bSuccess


endclass


///////////////////////////////////////////////
///  example:
///
///  oCgi = new CGIDownload()
///  oCgi.connect()
///  oCgi.Download("f:\data\dB2K.zip","dB2K.zip")
///  quit
///////////////////////////////////////////////

Class CGIDownload of CGISession from "WebClass.cc"

   function Download(cFilePath, cFileName)
      local oFile, nBytes,nTotBytes,cBytesRead
      oFile = new file()
      nBytes = oFile.size(cFilePath)
      oFile.open(cFilePath, "R")
      try
         this.fOut.puts("Content-Type: application/x-unknown")
         this.fOut.puts("Content-Disposition: attachment; filename="+cFilename)
         this.fOut.puts("")
         nTotBytes=0
         do 
        
            cBytesRead = oFile.Read(1024)
            this.fOut.write( cBytesRead )
            nTotBytes = nTotBytes + len(cBytesRead)

         until nTotBytes = nBytes

         oFile.close()
      catch (exception e)
         nBytes = 0
         try
            oFile.close()
         catch (exception e)
         endtry
      endtry

      return nBytes


   function DownloadTextFile( cFilePath, cFileName )

      local nLen, cText,oFile
      try
         oFile = new File()
         nLen = oFile.size( cFilePath )
         oFile.Open( cFilePath,"R")
         cText = ""
         do
            cText = cText + oFile.Read(256)
         until nLen = len(cText)
         oFile.Close()  

         this.fOut.Puts("Content-Type: application/x-unknown") 
         this.fOut.Puts("Content-Disposition: attachment; filename=" + cFileName) 
         this.fOut.Puts("")
         this.fOut.Puts(cEnv)

      catch (exception e)
         nLen = 0
         try
            oFile.close()
         catch (exception e)
         endtry

      endtry

      return nLen


endclass

